//package census;
import census.CensusObject;
import ij.*;
import ij.plugin.frame.PlugInFrame;
import ij.process.*;
import ij.gui.*;
import ij.io.*;
import java.awt.event.*;
import java.awt.Color;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;

import java.util.*;
import java.awt.*;
import ij_ImageCanvasWithOverlay.ImageCanvasWithOverlay;


/**
 * Census allows the user to mark the location of structures on
 * images and can apply plugins to further define the structure.  This
 * data can be saved and exported.
 *
 * Image object  are referred to as "sheep", the object types are referred to as "breeds".
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *
 * Peter Sebastian Masny
 * ps-contact@masny.dk
 *
 * Parts of this code were adapted from Wayne Rasband, Sergio
 * Caballero Jr., Michael A. Miller, and Philippe Thevenaz.  Content
 * derived from these authors follow their corresponding copyrights.
 *
 * @version    2.1 26 Jan 2009
 * @author     Peter Sebastian Masny
 */



/**
 *
 */
public class Census_ extends PlugInFrame implements ActionListener, MouseListener{
    
    
/*....................................................................
        Class variables
....................................................................*/
    Vector sheep = new Vector();
    String[] breedNames;
    static PluginClassLoader breedLoader;
    Vector breedClasses = new Vector();
    int[] colorBreed = {0,0,0,0,0,0,0};
    int[] showAs = {1,1,1,1,1,1,1};
    String[] showAsStrings = {"point","outline","fill"};
    boolean busy, listening;
    ImagePlus img;
    ImageCanvasWithOverlay icwo;
    String dataDir;
    String dataFile;
    String DATA_FILE_SUFFIX = "-censusData" + CensusObject.IO_DATA_SUFFIX;
    int[] colorTotals = {0,0,0,0,0,0,0};
    
/*....................................................................
        GUI components
....................................................................*/
    JToggleButton showpointJB,showoutlineJB,showfillJB;
    JToggleButton[] showColorJB = new JToggleButton[CensusObject.CHANNEL_COLOR[0].length];
    JToggleButton[] activeColorJB = new JToggleButton[CensusObject.CHANNEL_COLOR[0].length];
    JButton autofindJB,saveJB,loadJB,measureJB;
    JComboBox breedJCB;
    
/*....................................................................
        Census Constructor
....................................................................*/
    /*
     * This creates a new <code>Census</code> interface and displays it --
     * unless a <code>Census</code> exists, in which case this
     * constructor simply brings the previous one to the front, and
     * goes away.
     */
    public Census_() {
        super("Census");
        img = WindowManager.getCurrentImage();
        if (img == null) {
            IJ.error("No images open.\nPlease open images before starting Census.");
            return;
        }
        
        dataFile = img.getOriginalFileInfo().fileName + DATA_FILE_SUFFIX;
        dataDir =  img.getOriginalFileInfo().directory;
        //.replace('\\','/');
        // ImageJ seems to be OK with this in windows.  Allows cross platform consistency.
        
        // Load the user defined CensusObject extensions
        breedLoader = (PluginClassLoader)IJ.getClassLoader();
        try {
            if (loadBreed("census.CensusObject") == null) IJ.error("Unable to load CensusObject.class!\nCensus will not work.");
        } catch (ClassFormatError e) {
            IJ.error("ClassFormatError");
        }catch (NoClassDefFoundError e) {
            IJ.error("NoClassDefFoundError");
        }catch (SecurityException e) {
            IJ.error("SecurityException");
        }
        String censusDir = IJ.getDirectory("plugins")+"census";
        String[] censusDirFiles = (new File(censusDir)).list();
        Arrays.sort(censusDirFiles);
        ArrayList li = new ArrayList();
        for (int i =0; i< censusDirFiles.length;i++){
            if (censusDirFiles[i].toLowerCase().endsWith("breed.java")){
                boolean classExists = false;
                for (int j =0; j< censusDirFiles.length;j++) {
                    if (censusDirFiles[j].toLowerCase().equals(censusDirFiles[i].toLowerCase().replace("breed.java","breed.class"))) {
                        classExists = true;
                    }
                }
                if (!classExists) {
                    IJ.log("Compiling: "+censusDir+ File.separator + censusDirFiles[i]);
                    if (!breedCompile(censusDir+ File.separator + censusDirFiles[i])) {
                        IJ.error("Could not compile "+censusDirFiles[i]+"\nAborting");
                        break;
                    }
                    String type = (censusDirFiles[i].split("Breed.",2))[0];
                    Class iBreed = loadBreed("census."+type+"Breed");
                    if (iBreed == null) {
                        IJ.error("Failure loading breed class: "+type+"Breed");
                    } else {
                        li.add(type);
                        breedClasses.add(iBreed);
                    }
                }
            }
            if (censusDirFiles[i].toLowerCase().endsWith("breed.class")) {
                String type = (censusDirFiles[i].split("Breed.",2))[0];
                Class iBreed = loadBreed("census."+type+"Breed");
                if (iBreed == null) {
                    IJ.error("Failure loading breed class: "+type+"Breed");
                } else {
                    li.add(type);
                    breedClasses.add(iBreed);
                }
            }
        }
        breedNames = (String[])(li.toArray(new String[1]));
        icwo = new ImageCanvasWithOverlay(img);
        img.setWindow(new ImageWindow(img, icwo));
        
        makeWindow();
    }
    
    //    public void run(String arg) {} // Don't need this it seems
    
    
    
    static Class loadBreed(String className) {
        
        Class theClass = null;
        
        try {
            theClass = breedLoader.loadClass(className);
        } catch (ClassNotFoundException e) {
            if (className.indexOf('_')!=-1)
                IJ.error("Plugin not found: "+className);
        } catch (NoClassDefFoundError e) {
            if (className.indexOf('_')!=-1)
                IJ.error("Plugin not found: "+className);
        }
        return theClass;
    }
    
    
    
/*....................................................................
        breedCompile method
....................................................................*/
    /*
     * If a *breed.java file exists, and no corresponding .class
     * file exists, then this method compiles the .class file.
     * Code from ij.plugin.Compiler
     */
    static boolean breedCompile(String path) {
        sun.tools.javac.Main javac;
        ByteArrayOutputStream output = new ByteArrayOutputStream(4096);
        try {
             javac = new sun.tools.javac.Main(output, "javac");
        } catch (NoClassDefFoundError e) {
            IJ.error("This JVM does not include the javac compiler. Javac is\n"
                    +"included with the Windows, OS X and Linux versions of\n"
                    +"ImageJ.");
            return false;
        }
        
        
        IJ.showStatus("compiling: "+path);
        String classpath = System.getProperty("java.class.path");
        File f = new File(path);
        if (f!=null)  // add directory containing file to classpath
            classpath += File.pathSeparator + f.getParent();
        IJ.log("classpath: " + classpath);
        output.reset();
        String[] arguments;
        boolean compiled = javac.compile(new String[] {"-deprecation", "-classpath", classpath, path});
        String s = output.toString();
        boolean errors = s!=null && s.length()>0;
        if (errors && s.startsWith("Note: sun.tools.javac") && s.indexOf("error")==-1)
            errors = false;
        errors = (!compiled || errors);
        if (errors){
            IJ.error("Compile errors: "+s);
        } else
            IJ.showStatus("done");
        return compiled;
    }
/*....................................................................
        makeWindow method
....................................................................*/
    /*
     * Does a sanity check, then constructs the GUI and
     * attaches listener.
     */
    void makeWindow() {
        busy = true;

        
        img.getWindow().getCanvas().removeMouseListener(img.getWindow().getCanvas());
        img.getWindow().getCanvas().addMouseListener(this);
        listening = true;
        
        /*
         * Define buttons and button groups
         */
        for(int i=1; i< CensusObject.CHANNEL_COLOR[0].length; i++) {
            showColorJB[i] = new JToggleButton("<html><font color="+CensusObject.CHANNEL_COLOR[2][i]+">"+CensusObject.CHANNEL_COLOR[1][i]+": "+colorTotals[i]+"</font></html>");
            showColorJB[i].setSelected(true);
        }
        
        ButtonGroup activeBG = new ButtonGroup();
        for(int i=0; i< CensusObject.CHANNEL_COLOR[0].length; i++) {
            activeColorJB[i] = new JToggleButton(CensusObject.CHANNEL_COLOR[1][i]);
            activeBG.add(activeColorJB[i]);
        }
        activeColorJB[1].setSelected(true);
        
        ButtonGroup showBG = new ButtonGroup();
        showpointJB = new JToggleButton("point");
        showoutlineJB = new JToggleButton("outline");
        showfillJB = new JToggleButton("fill");
        showBG.add(showpointJB);
        showBG.add(showoutlineJB);
        showBG.add(showfillJB);
        showoutlineJB.setSelected(true);
        
        breedJCB = new JComboBox(breedNames);
        
        
        autofindJB = new JButton("Autofind");
        saveJB = new JButton("Save");
        loadJB = new JButton("Load");
        measureJB = new JButton("IJ Measure");
        
        
        /*
         * Attach the GUI listeners.
         */
        
        for(int i=1; i< CensusObject.CHANNEL_COLOR[0].length; i++) {
            showColorJB[i].addActionListener(this);
        }
        for (int i=0; i < CensusObject.CHANNEL_COLOR[0].length;i++) {
            activeColorJB[i].addActionListener(this);
        }
        showpointJB.addActionListener(this);
        showoutlineJB.addActionListener(this);
        showfillJB.addActionListener(this);
        breedJCB.addActionListener(this);
        
        autofindJB.addActionListener(this);
        saveJB.addActionListener(this);
        loadJB.addActionListener(this);
        measureJB.addActionListener(this);
        
        
        /*
         *  Construct the contentPane
         */
        JPanel contentPane = new JPanel();
        contentPane.setLayout( new BoxLayout(contentPane,BoxLayout.Y_AXIS));
        
        contentPane.add(new JLabel("Census window for: "+img.getOriginalFileInfo().fileName));
        
        JPanel colorsJP= new JPanel();
        colorsJP.add(new JLabel("Show colors:"));
        for(int i=1; i< CensusObject.CHANNEL_COLOR[0].length; i++) {
            colorsJP.add(showColorJB[i]);
        }
        contentPane.add(colorsJP);
        
        JPanel activeJP= new JPanel();
        activeJP.add(new JLabel("Active color:"));
        for(int i=0; i< CensusObject.CHANNEL_COLOR[0].length; i++) {
            activeJP.add(activeColorJB[i]);
        }
        contentPane.add(activeJP);
        
        JPanel showJP= new JPanel();
        showJP.add(new JLabel("Show object as:"));
        showJP.add(showpointJB);
        showJP.add(showoutlineJB);
        showJP.add(showfillJB);
        
        showJP.add(breedJCB);
        contentPane.add(showJP);
        
        JPanel fileJP = new JPanel();
        fileJP.add(autofindJB);
        fileJP.add(loadJB);
        fileJP.add(saveJB);
        fileJP.add(measureJB);
        contentPane.add(fileJP);
        
        
        contentPane.add(new JSeparator());
        contentPane.add(new JLabel("Click on image to mark. Ctrl-click to annotate. Shift-click to remove closest mark."));
        
        
        /*
         * Pack it up and present the interface for use
         */
        add(contentPane);
        pack();
        setVisible(true);
        busy = false;
    }
    
/*....................................................................
        rebuildOverlay method
....................................................................*/
    /*
     * Rebuilds the object overlays on all canvases
     */
    void rebuildOverlay() {
        icwo.overlayClear();
        ListIterator li = sheep.listIterator(0);
        while (li.hasNext()) {
            CensusObject next = (CensusObject)li.next();
            if(showColorJB[next.color].isSelected()){
                next.show(showAsStrings[showAs[next.color]]);
            }
        }
        icwo.repaint();
    }
    
/*....................................................................
        getActiveColor method
....................................................................*/
    /*
     * Reads the GUI to find the active color
     */
    
    int getActiveColor(){
        for (int i=0; i<CensusObject.CHANNEL_COLOR[0].length;i++){
            if (activeColorJB[i].isSelected()) return i;
        }
        return -1;
    }
    
    /*....................................................................
        fileLoad method
....................................................................*/
    /*
     * Loads objects from data file
     */
    
    void fileLoad(){
        busy = true;
        if (!(sheep.isEmpty())) {
            if (!(IJ.showMessageWithCancel("Census Load","You currently have spots marked.\nLoading will add objects, including\npossible overlap of identical objects\nDo you want to continue?"))) {
                return;
            }
        }
        try {
            BufferedReader br = new BufferedReader(new FileReader(dataDir+dataFile));
            String line = br.readLine();
            while (line != null) {
                String objectBreed = (line.split(CensusObject.IO_DELIMITER, -1))[2];
                int objectI = Arrays.binarySearch(breedNames,objectBreed);
                if (objectI >=0) {
                    CensusObject newObject = null;
                    try {
                        newObject = (CensusObject)((Class)(breedClasses.get(objectI))).newInstance();
                    } catch (InstantiationException ex) {
                        IJ.error("Unable to load plugin (ins)");
                    } catch (IllegalAccessException ex) {
                        IJ.error("Unable to load plugin, possibly \nbecause it is not public.");
                    }
                    boolean located = newObject.loadFromString(line, img);
                    if (located) {
                        sheep.add(newObject);
                        colorTotals[newObject.color]++;
                        showColorJB[newObject.color].setText("<html><font color="+CensusObject.CHANNEL_COLOR[2][newObject.color]+">"+CensusObject.CHANNEL_COLOR[1][newObject.color]+": "+colorTotals[newObject.color]+"</font></html>");
                    }
                }
                line = br.readLine();
            }
            br.close();
        } catch (IOException e) {
            IJ.showMessage("FileIO Error On Read: ", ""+e);
        }
        rebuildOverlay();
        busy = false;
    }
    
    /*....................................................................
        fileSave method
....................................................................*/
    /*
     * Save objects to data file.
     * Filename syntax: imagename + DataFileSuffix
     */
    
    void fileSave(){
        busy = true;
        try {
            if (new File(dataDir+dataFile).exists()) {
                if (!(IJ.showMessageWithCancel("Census Save","A data file already exists!\nDo you want to overwrite this file?"))) {
                    return;
                }
            }
            new File(dataDir+dataFile).delete();
            PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(dataDir+dataFile, true)));
            ListIterator li = sheep.listIterator(0);
            while (li.hasNext()) {
                pw.print(((CensusObject)(li.next())).asString()+"\n");
            }
            pw.flush();
            pw.close();
        } catch (IOException e) {
            IJ.showMessage("FileIO Error On Write: ", ""+e);
            return;
        }
        busy = false;
        IJ.showMessage("Census data file saved.");
    }
    
/*....................................................................
        actionPerformed method from ActionListener
....................................................................*/
    /*
     * On GUI buttonpresses that require action, performs that action.
     */
    public void actionPerformed(ActionEvent e){
        Object actionSource = e.getSource();
        
        for(int i=0; i< CensusObject.CHANNEL_COLOR[0].length; i++) {
            if (actionSource == activeColorJB[i]) {
                if (i==0){
                    img.getWindow().getCanvas().removeMouseListener(this);
                    img.getWindow().getCanvas().addMouseListener(img.getWindow().getCanvas());
                    listening = false;
                } else if (!listening) {
                    img.getWindow().getCanvas().removeMouseListener(img.getWindow().getCanvas());
                    img.getWindow().getCanvas().addMouseListener(this);
                    listening = true;
                }
                if (showAs[i]==0) {
                    showpointJB.setSelected(true);
                } else if (showAs[i] == 1) {
                    showoutlineJB.setSelected(true);
                } else {
                    showfillJB.setSelected(true);
                }
                breedJCB.setSelectedIndex(colorBreed[i]);
                return;
            }
        }
        
        for(int i=1; i< CensusObject.CHANNEL_COLOR[0].length; i++) {
            if (actionSource == showColorJB[i]) {
                rebuildOverlay();
                return;
            }
        }
        
        if (actionSource == measureJB) {
            ListIterator li = sheep.listIterator(0);
            while (li.hasNext()) {
                ((CensusObject)(li.next())).IJmeasure();
            }
        } else if  (actionSource == showpointJB) {
            showAs[getActiveColor()] = 0;
            rebuildOverlay();
        } else if  (actionSource == showoutlineJB) {
            showAs[getActiveColor()] = 1;
            rebuildOverlay();
        } else if  (actionSource == showfillJB) {
            showAs[getActiveColor()] = 2;
            rebuildOverlay();
        } else if  (actionSource == autofindJB) {
        } else if  (actionSource == saveJB) {
            fileSave();
        } else if  (actionSource == loadJB) {
            fileLoad();
        } else if  (actionSource == breedJCB) {
            colorBreed[getActiveColor()] = breedJCB.getSelectedIndex();
        }
    }
    
/*....................................................................
        ImageMouseListener extensions
....................................................................*/
    /*
     * Implementation of MouseListener that reacts to mousepresses
     * in the image windows.
     */
    
    public void mouseExited(final MouseEvent e){}
    public void mouseEntered(final MouseEvent e){}
    public void mouseReleased(final MouseEvent e){}
    public void mouseClicked(final MouseEvent e){}
    
    
    public void mousePressed(MouseEvent e){
            /*
             * Organize the mouseEvent data
             */
        ImageCanvas mouseSource = (ImageCanvas) e.getSource();
        int xMouse = mouseSource.offScreenX(e.getX());
        int yMouse = mouseSource.offScreenY(e.getY());
        boolean annotate = e.isControlDown();
        boolean delete = e.isShiftDown() && !(annotate);  //safety feature, annotate overrides.
        e.consume();
        
        if (busy) return;
        busy = true;
        
        int activeColor = getActiveColor();
        if (activeColor <1 ) {
            busy = false;
            // doParentlistener action...all this shouldn't trigger
            return;
        }
        
        if (delete) {
            if (sheep.isEmpty()) {
                IJ.error("No objects to delete.");
            } else {
                int deleteMe = -1;
                double closestDistance = Double.POSITIVE_INFINITY;
                ListIterator li = sheep.listIterator(0);
                while (li.hasNext()) {
                    CensusObject current = (CensusObject)li.next();
                    if (current.color == activeColor) {
                        float distance2 = (xMouse-current.x)*(xMouse-current.x)+(yMouse-current.y)*(yMouse-current.y);
                        if ( distance2 < closestDistance) {
                            deleteMe = li.previousIndex();
                            closestDistance = distance2;
                        }
                    }
                }
                if (deleteMe > -1) {
                    sheep.remove(deleteMe);
                    colorTotals[activeColor]--;
                    showColorJB[activeColor].setText("<html><font color="+CensusObject.CHANNEL_COLOR[2][activeColor]+">"+CensusObject.CHANNEL_COLOR[1][activeColor]+": "+colorTotals[activeColor]+"</font></html>");
                } else {
                    IJ.error("No "+CensusObject.CHANNEL_COLOR[0][activeColor]+" objects to delete.");
                }
                rebuildOverlay();
            }
        } else if (annotate) {
        if (sheep.isEmpty()) {
                IJ.error("No objects to annotate.");
            } else {
                CensusObject annoMe = null;
                double closestDistance = Double.POSITIVE_INFINITY;
                ListIterator li = sheep.listIterator(0);
                while (li.hasNext()) {
                    CensusObject current = (CensusObject)li.next();
                    if (current.color == activeColor) {
                        float distance2 = (xMouse-current.x)*(xMouse-current.x)+(yMouse-current.y)*(yMouse-current.y);
                        if ( distance2 < closestDistance) {
                            annoMe = current;
                            closestDistance = distance2;
                        }
                    }
                }
                if (annoMe != null) {
                    annoMe.annotate();
                } else {
                    IJ.error("No "+CensusObject.CHANNEL_COLOR[0][activeColor]+" objects to annotate.");
                }
                rebuildOverlay();
            }
        } else {
            CensusObject newObject = null;
            int objectI = breedJCB.getSelectedIndex();
            try {
                newObject = (CensusObject)((Class)(breedClasses.get(objectI))).newInstance();
            } catch (InstantiationException ex) {
                IJ.error("Unable to load plugin (ins)");
            } catch (IllegalAccessException ex) {
                IJ.error("Unable to load plugin, possibly \nbecause it is not public.");
            }
            boolean located = newObject.find(xMouse,yMouse,img,activeColor);
            if (located) {
                sheep.add(newObject);
                if (showColorJB[activeColor].isSelected()) {
                    newObject.show(showAsStrings[showAs[activeColor]]);
                }
                icwo.repaint();
                colorTotals[activeColor]++;
                showColorJB[activeColor].setText("<html><font color="+CensusObject.CHANNEL_COLOR[2][activeColor]+">"+CensusObject.CHANNEL_COLOR[1][activeColor]+": "+colorTotals[activeColor]+"</font></html>");
            }
        }
        busy = false;
    }
}


